//GAMY: a tool for estimating the similarity of graphs
//	Copyright(C) 2019 Mattia De Prisco and Giovanni Squillero
//
//	This program is free software : you can redistribute it and /or modify
//	it under the terms of the GNU General Public License as published by
//	the Free Software Foundation, either version 3 of the License, or
//	(at your option) any later version.
//
//	This program is distributed in the hope that it will be useful,
//	but WITHOUT ANY WARRANTY; without even the implied warranty of
//	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.See the
//	GNU General Public License for more details.
//
//	You should have received a copy of the GNU General Public License
//	along with this program.If not, see < http://www.gnu.org/licenses/>.

#include "graph.h"
#include <string>
#include <regex>
#include <fstream>
#include <sstream>

bool Graph::VERTEX_LABELS_ON = true;

// static inits can't be .h
std::vector<Node*> NodeStore::nodes_ = std::vector<Node*>();

inline Node* Graph::addNode(const std::string& name, const std::string& label) {
	nodes_.push_back(NodeStore::newNode(name, label));
	return nodes_.back();
}

inline void Graph::addEdge(Node* from, Node* to) {
	from->to_.push_back(to);
	to->from_.push_back(from);
}

Graph::Graph(const std::string& filename) {
	filename_ = filename.substr(filename.rfind('/') + 1);
	const auto extension = filename.substr(filename.rfind('.') + 1);
	read(filename, extension);
}

std::vector<node_t> Graph::nodeList() const {
	std::vector<node_t> lst;
	for (auto n : nodes_)
		lst.push_back(n->id_);
	return lst;
}

void Graph::read(const std::string & filename, const std::string & extension) {
	if (extension == "mat")
		readMat(filename);
	else if (extension == "ladder")
		readLadder(filename);
	else if (extension == "dimacs")
		readDimacs(filename);
	else if (extension == "yara")
		readYara(filename);
	else
		throw GraphException("Invalid graph format.");
}

void Graph::readMat(const std::string & filename) {

	std::ifstream file(filename);
	if (!file.is_open())
		throw GraphException("Error while opening file " + filename);


	std::getline(file, md5_);
	std::getline(file, application_);

	node_t n_nodes;
	file >> n_nodes;

	std::string garbage;
	std::getline(file, garbage);

	for (auto t = 0; t < n_nodes; ++t) {
		std::string line;
		std::getline(file, line);
		std::string::iterator it;
		for (it = line.end() - 1; *it != ' '; --it);
		const std::string name(line.begin(), it);
		const std::string label(it + 1, line.end());
		if (VERTEX_LABELS_ON) {
			addNode(name, label);
		} else {
			addNode(name, "none");
		}
	}


	for (node_t n1 = 0; n1 < n_nodes; n1++) {
		std::string line;
		std::getline(file, line);
		node_t n2 = 0;
		for (const auto& c : line) {
			if (c == '1') {
				addEdge(nodes_.at(n1), nodes_.at(n2));
			}
			if (c != ' ')
				n2++;
		}
	}
	file.close();

}

void Graph::readLadder(const std::string & filename) {
	std::ifstream file(filename);
	if (!file.is_open())
		throw GraphException("Error while opening file " + filename);

	int n_nodes;
	file >> n_nodes;

	for (int i = 0; i < n_nodes; i++)
		addNode("name", "label");

	for (int i = 0; i < n_nodes; i++) {
		int edge_count;
		file >> edge_count;
		for (int j = 0; j < edge_count; j++) {
			int w;
			file >> w;
			addEdge(nodes_.at(i), nodes_.at(w));
		}
	}
	file.close();
}

void Graph::readDimacs(const std::string & filename) {
	std::ifstream file(filename);
	if (!file.is_open())
		throw GraphException("Error while opening file " + filename);

	std::string line;
	while (std::getline(file, line)) {
		switch (line.front()) {
		case 'e':
		{
			int from_node, to_node;
			std::string temp;
			std::stringstream(line) >> temp >> from_node >> to_node;
			addEdge(nodes_.at(from_node - 1), nodes_.at(to_node - 1));
			break;
		}
		case 'n':
		{
			int name, label;
			std::string temp;
			std::stringstream(line) >> temp >> name >> label;
			if (VERTEX_LABELS_ON) {
				addNode(std::to_string(name), std::to_string(label));
			} else {
				addNode("name", "none");
			}
		}
		default:
			break;

		}
	}
	file.close();
}

void Graph::readYara(const std::string& filename) {
	std::ifstream file(filename);
	if (!file.is_open())
		throw GraphException("Error while opening file " + filename);

	node_t n_nodes;
	file >> n_nodes;
	for(node_t i=0; i < n_nodes; i++) {
		std::string name, label;
		file >> name >> label;
		if (VERTEX_LABELS_ON) {
			addNode(name, label);
		} else {
			addNode(name, "none");
		}
	}

	std::string line;
	std::getline(file, line);

	for (node_t n1 = 0; n1 < n_nodes; n1++) {
		std::getline(file, line);
		node_t n2 = 0;
		for (const auto& c : line) {
			if (c == '1') {
				addEdge(nodes_.at(n1), nodes_.at(n2));
			}
			if (c != ' ')
				n2++;
		}
	}
	file.close();
}