//GAMY: a tool for estimating the similarity of graphs
//	Copyright(C) 2019 Mattia De Prisco and Giovanni Squillero
//
//	This program is free software : you can redistribute it and /or modify
//	it under the terms of the GNU General Public License as published by
//	the Free Software Foundation, either version 3 of the License, or
//	(at your option) any later version.
//
//	This program is distributed in the hope that it will be useful,
//	but WITHOUT ANY WARRANTY; without even the implied warranty of
//	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.See the
//	GNU General Public License for more details.
//
//	You should have received a copy of the GNU General Public License
//	along with this program.If not, see < http://www.gnu.org/licenses/>.

#pragma once

#include <string>
#include <utility>
#include <vector>

// Node index
typedef int_fast32_t node_t;

// paranoia checks
static_assert(std::is_signed<node_t>::value, "node_t must be signed");
static_assert(std::is_integral<node_t>::value, "node_t must be integer");
//static_assert(std::numeric_limits<node_t>::max() >= std::numeric_limits<size_t>::max(), "node_t can't be narrower than size_t");
//static_assert(sizeof(node_t) >= sizeof(size_t), "node_t can't be narrower than size_t");

// forward declarations
class NodeStore;
class Graph;

//
// Simplistic class to handle exceptions
//
struct GraphException : public std::exception {
    std::string good_reason;
    explicit GraphException() : good_reason{"Panic!"} {}
    explicit GraphException(std::string exp) : good_reason{std::move(exp)} {}
};

//
// A GetNode in a graph
//
class Node {
    friend class NodeStore;
    friend class Graph;

    std::vector<Node*> from_;
    std::vector<Node*> to_;
    node_t id_;

    Node(node_t id) : id_{id}, label{""} {};
    Node(node_t id, std::string label) : id_{id}, label{std::move(label)} {};
public:

	 Node() = delete;

    std::string label;
    std::string name;
    operator node_t() const {
        return id_;
    }
    size_t inDegree() const {
        return from_.size();
    }
    size_t outDegree() const {
        return to_.size();
    }
    size_t degree() const {
        return inDegree() + outDegree();
    }
    std::vector<node_t> from() const {
        auto ret = std::vector<node_t>(from_.size());
        for(size_t i = 0; i < from_.size(); ++i)
            ret[i] = from_[i]->id_;
        return ret;
    };
    std::vector<node_t> to() const {
        auto ret = std::vector<node_t>(to_.size());
        for(size_t i = 0; i < to_.size(); ++i) {
            ret[i] = to_[i]->id_;
        }
        return ret;
    };
};

//
// Node_store: a global vector of nodes to keep int<->Nodes with the Hyper partitions
//
class NodeStore {
    static std::vector<Node*> nodes_;
public:
    static Node* newNode() {
        nodes_.push_back(new Node(static_cast<node_t>(nodes_.size())));    // can't use emplace_back w/ private constructor!
        return nodes_.back();
    }
    static Node* newNode(const std::string& name, const std::string& label) {
        nodes_.push_back(new Node(static_cast<node_t>(nodes_.size()), label));    // can't use emplace_back w/ private constructor!
        return nodes_.back();
    }

    static Node*& getNode(const node_t index) {
//#ifdef NDEBUG
        return nodes_[index];
//#else
//        return nodes.at(index);
//#endif // NDEBUG
    }
    static size_t size() { return nodes_.size(); }

	static std::vector<Node*>& getNodes() {
		return nodes_;
    }
};


//
// A Graph
//
class Graph {

    std::vector<Node*> nodes_;
    std::string md5_;
    std::string application_ = "NONE";
	std::string filename_;

	void read(const std::string& filename, const std::string& extension);
    void readMat(const std::string& filename);
	void readLadder(const std::string& filename);
	void readDimacs(const std::string& filename);
	void readYara(const std::string& filename);
    
    inline Node* addNode(const std::string& name, const std::string& label);
    inline void addEdge(Node* from, Node* to);


public:
	static bool VERTEX_LABELS_ON;

	Graph(const std::string& filename);
    
	std::vector<node_t> nodeList() const;
	const std::string& getApplicationName() const { return application_; };
	const std::string& getFileName() const { return filename_; };
	const std::string& getMd5() const { return md5_; }
	const std::vector<Node*>& getNodes() const { return nodes_; };
	size_t size() const { return nodes_.size(); };
};
