//GAMY: a tool for estimating the similarity of graphs
//	Copyright(C) 2019 Mattia De Prisco and Giovanni Squillero
//
//	This program is free software : you can redistribute it and /or modify
//	it under the terms of the GNU General Public License as published by
//	the Free Software Foundation, either version 3 of the License, or
//	(at your option) any later version.
//
//	This program is distributed in the hope that it will be useful,
//	but WITHOUT ANY WARRANTY; without even the implied warranty of
//	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.See the
//	GNU General Public License for more details.
//
//	You should have received a copy of the GNU General Public License
//	along with this program.If not, see < http://www.gnu.org/licenses/>.

#include "logy.h"
#include "graph.h"
#include "hyperpartition.h"
#include <algorithm>
#include <random>


//#include "graph.h"
double PartitionManager::EQUIVALENCE_MODIFIER = 2.0;
double PartitionManager::HOPE_REDUCER = 0.8;
constexpr node_t PartitionManager::CHAIN_END;

bool PartitionManager::consistencyCheck() const {
    for(auto d : data_) {
        if(d != PartitionManager::ERASED && d != PartitionManager::CHAIN_END && d >= static_cast<node_t>(data_.size())) {
            LOGY_WARNING("Illegal value:", d);
            return false;
        }
    }

    for(auto hp : hyper_partitions_) {
        if(hp.partition_heads.max_size() == 0) {
            LOGY_WARNING("Empty hyper partition");
            return false;
        }
        if(hp.partition_heads.size() != hyper_partitions_[0].partition_heads.size()) {
            LOGY_WARNING("Inconsisten size");
            return false;
        }
    }

    std::vector<size_t> counter(data_.size(), 0);
    for(size_t i = 0; i < data_.size(); ++i)
        if(data_[i] == PartitionManager::ERASED)
            counter[i] = 1;
    for(auto hp : hyper_partitions_) {
        for(auto p : hp.partition_heads) {
            while(p != PartitionManager::CHAIN_END) {
                if(++counter[p] > 1) {
                    LOGY_WARNING("Doubly linked element");
                    return false;
                }
                p = data_[p];
            }
        }
    }
    if(std::find(counter.begin(), counter.end(), 0) != counter.end()) {
        LOGY_WARNING("Missing element");
        return false;
    }

    return true;
}

// Sort partitions according to some criteria
void PartitionManager::sort(const SortStrategy strategy) {
    for(size_t i = 0; i < hyper_partitions_.size(); ++i)
        sort(i, strategy);
}

void PartitionManager::sort(const size_t partition_index, const SortStrategy strategy) {
	for (node_t& partition_head : hyper_partitions_[partition_index].partition_heads) {
        auto e = partition_head;
        if(e == PartitionManager::CHAIN_END)
            continue;

        std::vector<node_t> partition;
        while(e != PartitionManager::CHAIN_END) {
            partition.push_back(e);
            e = data_[e];
        }
        //LOG_DEBUG("OLD Partition", i, ":", partition);

        // sort the partition
        switch(strategy) {
        case PartitionManager::SortStrategy::NATURAL:
            std::sort(partition.begin(), partition.end());
            break;
        case PartitionManager::SortStrategy::DEGREE:
            std::sort(partition.begin(), partition.end(),
                [](node_t a, node_t b) {return NodeStore::getNode(a)->degree() > NodeStore::getNode(b)->degree(); });
            break;
        case PartitionManager::SortStrategy::IN_DEGREE:
            std::sort(partition.begin(), partition.end(),
                [](node_t a, node_t b) {return NodeStore::getNode(a)->inDegree() > NodeStore::getNode(b)->inDegree(); });
            break;
        case PartitionManager::SortStrategy::OUT_DEGREE:
            std::sort(partition.begin(), partition.end(),
                [](node_t a, node_t b) {return NodeStore::getNode(a)->outDegree() > NodeStore::getNode(b)->outDegree(); });
            break;
        case PartitionManager::SortStrategy::RANDOM:
            std::shuffle(partition.begin(), partition.end(), std::default_random_engine{});
            break;
        default:
            break;
        }

        // save back the partition
        partition_head = partition[0];
        auto itr = partition.begin();
        while(itr + 1 != partition.end()) {
            data_[*itr] = *(itr + 1);
            ++itr;
        }
        data_[*itr] = PartitionManager::CHAIN_END;
        //LOG_DEBUG("NEW Partition", i, ":", partition);
    }

    //assert(consistency_check());
}

PartitionManager::PartitionManager(std::vector<std::vector<node_t>> partitions) {
    size_t size = 0;
    for(const auto& p : partitions)
        size += p.size();
    data_ = std::vector<node_t>(size, -2);
	Partition part0;
	part0.partition_heads = std::vector<node_t>(partitions.size());
    for(size_t p = 0; p < partitions.size(); ++p) {
        part0.partition_heads[p] = partitions[p][0];
        for(size_t q = 0; q < partitions[p].size() - 1; ++q)
//#ifdef NDEBUG
            data_[partitions[p][q]] = partitions[p][q + 1];
//#else
//            data.at(partitions.at(p).at(q)) = partitions.at(p).at(q + 1);
//#endif
        data_[partitions[p].back()] = PartitionManager::CHAIN_END;
    }
    hyper_partitions_.push_back(part0);

//#ifndef NDEBUG
//    if(std::find(data.begin(), data.end(), -2) != data.end())
//        LOGY_WARNING("Some elements have not been initialized.");
//#endif

    //assert(consistency_check());
}

PartitionManager::PartitionManager(PartitionManager&& other) noexcept : data_(std::move(other.data_)),
	hyper_partitions_(std::move(other.hyper_partitions_)) {};

PartitionManager& PartitionManager::operator=(PartitionManager&& other) noexcept {
	data_ = std::move(other.data_);
	hyper_partitions_ = std::move(other.hyper_partitions_);
	return *this;
};

void PartitionManager::dump() const {
    LOGY("Found", hyper_partitions_.size(), "hyper partitions");
    std::vector<node_t> erased;
    for(size_t n = 0; n < data_.size(); ++n)
        if(data_[n] == PartitionManager::ERASED)
            erased.push_back(static_cast<int>(n));
    LOGY("*", ">", erased);
    for(size_t c = 0; c < hyper_partitions_.size(); ++c) {
		LOGY(c, ">", getHyperPartitions(c));
		LOGY("hope:", partitionHope(c));
    }
    //assert(consistency_check());
}

//
// Splits the current partitions by intersecting them with a set
//
void PartitionManager::split(const std::vector<node_t>& list) {
    std::unordered_set<node_t> set;
    for(const auto& n : list)
        set.insert(n);
    split(set);
}

void PartitionManager::split(const std::unordered_set<node_t>& nodes_to_split) {
//#ifndef NDEBUG
//    for(auto n : set)
//        if(n < 0 || n >= data.size())
//            LOGY_WARNING("Can't find element", n);
//#endif
	auto new_partitions = std::vector<Partition>();
	for (auto& hyper_partition : hyper_partitions_) {
		Partition new_pset;
		new_pset.partition_heads = std::vector<node_t>(hyper_partition.partition_heads.size(), PartitionManager::CHAIN_END);
		for(size_t current_head = 0; current_head < hyper_partition.partition_heads.size(); ++current_head) { 
			const auto head_present = nodes_to_split.find(hyper_partition.partition_heads[current_head]) != nodes_to_split.end();
			node_t* current_node = &hyper_partition.partition_heads[current_head];             
			while(*current_node != PartitionManager::CHAIN_END) {
				if(head_present && nodes_to_split.find(*current_node) == nodes_to_split.end() ||
				   !head_present && nodes_to_split.find(*current_node) != nodes_to_split.end()) {
					// if partition head IS in set but current node IS NOT
					// or partition head IS NOT in set but current node IS
					// insert the current node in the new partition
					const auto tmp = data_[*current_node];
					data_[*current_node] = new_pset.partition_heads[current_head];
					new_pset.partition_heads[current_head] = *current_node;
					*current_node = tmp;
				} else {
					current_node = &data_[*current_node];
				}
			}
			if(head_present) 
				std::swap(new_pset.partition_heads[current_head], hyper_partition.partition_heads[current_head]);
		}

		new_pset.partition_hope = partitionHope(new_pset.partition_heads);
		hyper_partition.partition_hope = partitionHope(hyper_partition.partition_heads);

		if(maxPartitionSize(hyper_partition.partition_heads) == 0)
			// original partition has been exhausted
			hyper_partition = std::move(new_pset);
		else if(maxPartitionSize(new_pset.partition_heads) > 0)
			// both partitions are required
			new_partitions.push_back(std::move(new_pset));
		//else
			// the new partition is empty, we can safely discard it
			//assert(max_partition_size(new_pset.partition_heads) == 0); // do nothing
	}
	
	//assert(max_partition_size(new_pset.partition_heads) == 0);

	std::move(new_partitions.begin(), new_partitions.end(), std::back_inserter(hyper_partitions_));
	std::sort(hyper_partitions_.begin(), hyper_partitions_.end(),
			  [](const Partition& a, const Partition& b) { return a.partition_hope > b.partition_hope; });
}

size_t PartitionManager::chainLength(node_t element) const {
    size_t len = 0;
    while(element != PartitionManager::CHAIN_END) {
        ++len;
        element = data_[element];
    }
    return len;
}

double PartitionManager::partitionHope(const size_t& hpart_index) const {
    return partitionHope(hyper_partitions_[hpart_index].partition_heads);
}

// possible solutions given the hyper partition
double PartitionManager::partitionHope(const std::vector<node_t>& hyper_partition) const {

	double hope = 0;
	size_t max_eq_length = 0;
	
	std::vector<size_t> chain_lengths;
	for(const auto& head : hyper_partition) {
		auto len = chainLength(head);
		if(len > 0) {
			chain_lengths.push_back(len);
			max_eq_length++;
		}
			
	}
	if (chain_lengths.size() < 2)
		return hope;
	std::sort(chain_lengths.begin(), chain_lengths.end(),
		[](const size_t& a, const size_t& b) {return a < b; });

	size_t current_min_length = chain_lengths.front();
	int current_pos = 0;
	while (max_eq_length > 1) {

		/*MATH VERSION*/
		if (current_min_length != 0) {
			double full_value = 1;
			for (size_t i = 0; i < max_eq_length; i++) {
				full_value *= EQUIVALENCE_MODIFIER;
			}

			double res = 1;
			for (size_t i = 0; i < current_min_length; i++) {
				res *= HOPE_REDUCER;
			}

			hope += full_value * ((1 - res) / (1 - HOPE_REDUCER));
		}

		/* DESCRIPTIVE VERSION */

		//double modifier = 1;
		//for(size_t i = 0; i < current_min_length; i++) {
		//	double result = 1;
		//	for(size_t j = 0; j < max_eq_length; j++) {
		//		result *= EQUIVALENCE_MODIFIER;
		//	}
		//	hope += result * modifier;
		//	modifier *= PART_HOPE_REDUCER;
		//}

		current_pos++;

		if (current_min_length != 0) {
			std::transform(chain_lengths.begin() + current_pos, chain_lengths.end(), chain_lengths.begin() + current_pos,
				[&current_min_length](size_t d) -> std::size_t { return d - current_min_length; });
		}
		current_min_length = chain_lengths.at(current_pos);
		max_eq_length--;
	}
	return hope;
}

size_t PartitionManager::minPartitionSize(std::vector<node_t> hpart) const {
    size_t sz = chainLength(hpart[0]);
    for(size_t i = 1; i < hpart.size(); ++i)
        sz = std::min(sz, chainLength(hpart[i]));
    return sz;
}

size_t PartitionManager::maxPartitionSize(std::vector<node_t> heads) const {
    size_t sz = chainLength(heads[0]);
    for(size_t i = 1; i < heads.size(); ++i)
        sz = std::max(sz, chainLength(heads[i]));
    return sz;
}

// returns the current partitions as a vector of partitions, each partition contains a vector of graph,
// each graph contains a vecto of nodes
std::vector<std::vector<std::vector<node_t>>> PartitionManager::getHyperPartitions() const {
    std::vector<std::vector<std::vector<node_t>>> hpart;

    for(const auto& hp : hyper_partitions_) {
        std::vector<std::vector<node_t>> pset;
        for(node_t p : hp.partition_heads) {
            std::vector<node_t> part;
            while(p != PartitionManager::CHAIN_END) {
                part.push_back(p);
                p = data_[p];
            }
            pset.push_back(part);
        }
        hpart.push_back(pset);
    }

   // assert(consistency_check());
    return hpart;
}

// returns the current partitions as a vector of partitions, each partition contains a vector of graph,
// each graph contains a vecto of nodes
std::vector<std::vector<node_t>> PartitionManager::getHyperPartitions(const size_t partition_index) const {
    std::vector<std::vector<node_t>> hpart;

    for(node_t p : hyper_partitions_[partition_index].partition_heads) {
        std::vector<node_t> part;
        while(p != PartitionManager::CHAIN_END) {
            part.push_back(p);
            p = data_[p];
        }
    	hpart.push_back(part);
    }

   //assert(consistency_check());
    return hpart;
}

// remove a node_t from the partition
void PartitionManager::forget(const node_t node) {
    LOGY_DEBUG("Forget about", node);
    //assert(data[node] != Partition_manager::ERASED);

    const auto parent_data = std::find(data_.begin(), data_.end(), node);
    if(parent_data != data_.end()) {
        // found in data
        LOGY_DEBUG("DATA:", data);
        *parent_data = data_[node];
        data_[node] = PartitionManager::ERASED;
        LOGY_DEBUG("DATA:", data);
    } else {
        // it's a head!
        for(auto& hp : hyper_partitions_) {
            auto parent_head = std::find(hp.partition_heads.begin(), hp.partition_heads.end(), node);
            if(parent_head != hp.partition_heads.end()) {
                LOGY_DEBUG("HEAD:", hp.partition_heads);
                *parent_head = data_[*parent_head];
                data_[node] = PartitionManager::ERASED;
                LOGY_DEBUG("HEAD:", hp.partition_heads);
                LOGY_DEBUG("DATA:", data);
                break;
            }
        }
    }
    //assert(consistency_check());
}