//GAMY: a tool for estimating the similarity of graphs
//	Copyright(C) 2019 Mattia De Prisco and Giovanni Squillero
//
//	This program is free software : you can redistribute it and /or modify
//	it under the terms of the GNU General Public License as published by
//	the Free Software Foundation, either version 3 of the License, or
//	(at your option) any later version.
//
//	This program is distributed in the hope that it will be useful,
//	but WITHOUT ANY WARRANTY; without even the implied warranty of
//	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.See the
//	GNU General Public License for more details.
//
//	You should have received a copy of the GNU General Public License
//	along with this program.If not, see < http://www.gnu.org/licenses/>.

#pragma once

#include "graph.h"
#include <vector>
#include <unordered_set>

class PartitionManager {
public:
    enum class SortStrategy { NATURAL, DEGREE, IN_DEGREE, OUT_DEGREE, RANDOM };
	static constexpr node_t CHAIN_END = -1;
    static constexpr node_t ERASED = -2;

	static double EQUIVALENCE_MODIFIER;
	static double HOPE_REDUCER;

private:

    std::vector<node_t> data_;
	struct Partition
	{
		std::vector<node_t> partition_heads;
		double partition_hope{0};

		Partition() = default;
		~Partition() = default;
		Partition(const Partition& other) =	default;
		Partition& operator=(const Partition& other) =	default;
		
		Partition(Partition&& other) noexcept : partition_heads(std::move(other.partition_heads)), partition_hope(other.partition_hope) {};
		Partition& operator=(Partition&& other) noexcept {
			partition_heads = std::move(other.partition_heads);
			partition_hope = other.partition_hope;
			return *this;
		};
	};
	std::vector<Partition> hyper_partitions_;   

    size_t minPartitionSize(std::vector<node_t> hpart) const;
    size_t maxPartitionSize(std::vector<node_t> heads) const;

    bool consistencyCheck() const;

public:
    PartitionManager() = delete;
	~PartitionManager() = default;
    explicit PartitionManager(std::vector<std::vector<node_t>> partitions);
	
	PartitionManager(const PartitionManager& other) = default;
	PartitionManager& operator=(const PartitionManager& other) = default;
	
	PartitionManager(PartitionManager&& other) noexcept;
	PartitionManager& operator=(PartitionManager&& other) noexcept;

    size_t size() const { return hyper_partitions_.size(); }

    std::vector<size_t> size(size_t hyper_partition_index) const {
        std::vector<size_t> r(hyper_partitions_[hyper_partition_index].partition_heads.size());
        for(size_t i = 0; i < hyper_partitions_[hyper_partition_index].partition_heads.size(); ++i)
            r[i] = chainLength(hyper_partitions_[hyper_partition_index].partition_heads[i]);
        return r;
    }

    double partitionHope(const size_t& hpart_index) const;
    double partitionHope(const std::vector<node_t>& hyper_partition) const;
	size_t chainLength(node_t element) const;

	const std::vector<Partition>& getHp() const { return hyper_partitions_; } ;

    // Split hyper partitions according to a given set
    void split(const std::unordered_set<node_t>& nodes_to_split);
    void split(const std::vector<node_t>& list);
    
	// Unlink from all partitions (ie. "forget" about it)
    void forget(const node_t node);

    // Reorder partition according to some criteria
    void sort(const SortStrategy strategy);
    void sort(const size_t partition_index, const SortStrategy strategy);

    // Get partitions
    std::vector<std::vector<std::vector<node_t>>> getHyperPartitions() const;
    std::vector<std::vector<node_t>> getHyperPartitions(const size_t partition_index) const;

    // debug print
    void dump() const;
};
