//GAMY: a tool for estimating the similarity of graphs
//	Copyright(C) 2019 Mattia De Prisco and Giovanni Squillero
//
//	This program is free software : you can redistribute it and /or modify
//	it under the terms of the GNU General Public License as published by
//	the Free Software Foundation, either version 3 of the License, or
//	(at your option) any later version.
//
//	This program is distributed in the hope that it will be useful,
//	but WITHOUT ANY WARRANTY; without even the implied warranty of
//	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.See the
//	GNU General Public License for more details.
//
//	You should have received a copy of the GNU General Public License
//	along with this program.If not, see < http://www.gnu.org/licenses/>.

#include "logy.h"
#include "montecarlonode.h"
#include "montecarlotree.h"
#include "selector.h"
#include "history.h"
#include <algorithm>
#include <fstream>
#include <utility>

double MonteCarloNode::NODE_HOPE_MODIFIER = 1.0;
double MonteCarloNode::FACT_MODIFIER = 10.0;
double MonteCarloNode::PROPAGATION_INITIAL_VALUE = 1.0;
double MonteCarloNode::PROPAGATION_MODIFIER = 0.8;

bool MonteCarloNode::findNewEquivalence(const std::weak_ptr<MonteCarloNode>& weak_this) {
	// step 1: choose partition
	Selector<size_t, double> partition_selector;
	for(size_t pi = 0; pi < partitions_.size(); ++pi) {
		LOGY_DEBUG(partitions.size(pi), "->", "hope:", partitions.partition_hope(pi), "value:", Partition_value(pi));
		partition_selector.addEntry(pi, partitionValue(pi));
	}
	// vanilla roulete_wheel -> partitions with a value of 0 can't be selected
	while(partition_selector.size() > 0) {
		const auto selected_partition = partition_selector.rouletteWheel(); //Partition removed after being chosen

		LOGY_DEBUG("Selected partition:", selected_partition);

		// step 2: select the expansion
		auto hyper_partition = partitions_.getHyperPartitions(selected_partition);
		LOGY_DEBUG("hyper_partition:", hyper_partition);

		while(true) {
			std::vector<node_t> new_equivalence;
			for(auto& partition : hyper_partition) {
				if(!partition.empty()) {
					// favor low index, i.e. potentially useful nodes (the partition is sorted)
					const size_t r = SelectorUtilities::tournamentRangeInverted(partition.size(), 2);
					LOGY_DEBUG(partition, "=[", r, "]=>", partition[r]);
					new_equivalence.push_back(partition[r]);
					partition.erase(partition.begin() + r); //Node removed after being chosen
				} else {
					LOGY_DEBUG("Empty partition");
				}
			}

			if(new_equivalence.size() < 2)
				break;

			History new_history = getHistory();
			new_history.addChoice(new_equivalence);

			if(tree_->findNode(new_history))
				tree_->helper.avoided++;
			else {
				// step 3: do the actual expansion
				expand(std::move(new_history), new_equivalence, weak_this);
				return true;
			}		
		}
	}
	//Node completely visited -> remove it from candidates
	tree_->helper.closed_by_exausthion++;
	return false;
}

void MonteCarloNode::expand(History&& new_history, const std::vector<node_t>& new_equivalence, const std::weak_ptr<MonteCarloNode>& weak_this) {
	LOGY_DEBUG("Expanding:", new_equivalence);
	std::vector<node_t> from;
	std::vector<node_t> to;
	PartitionManager child_partitions = partitions_;
	for(auto n : new_equivalence) {
		child_partitions.forget(n);
	}
	for(const auto node_i : new_equivalence) {
		const auto node = NodeStore::getNode(node_i);
		auto f = node->from();
		std::copy(f.begin(), f.end(), std::back_inserter(from));
		auto t = node->to();
		std::copy(t.begin(), t.end(), std::back_inserter(to));
		LOGY_DEBUG("From:", f, "To:", t);
	}
	LOGY_DEBUG("FROM->", from);
	LOGY_DEBUG("TO->", to);

	child_partitions.split(from);
	//child_partitions.dump();
	child_partitions.split(to);
	//child_partitions.dump();

	auto new_node = std::make_shared<MonteCarloNode>(weak_this, std::move(child_partitions), new_history);
	new_node->backpropagate();
	if(!tree_->badWeight(new_node->weight_, new_node->fact_, true)) {
		std::weak_ptr<MonteCarloNode> ref = new_node;
		children_.emplace_back(new_node);
		tree_->updateBestSolution(new_node);
		if (new_node->hope_ != 0) {
			tree_->insertNewCandidate(std::move(ref));
		} else {
			tree_->helper.closed_by_exausthion++;
		}
		tree_->insertNewNode(std::move(new_history), std::move(new_node));
	} else {
		tree_->helper.closed_new++;
	}
}

void MonteCarloNode::printEquivalences(std::ofstream& of) const {
	for(const auto& equivalence : getHistory().getChoices()) {
			of << "( ";
			for(size_t i = 0; i < equivalence.size(); i++) {
				if(i + 1 < equivalence.size())
					of << equivalence.at(i) << " - ";
				else
					of << equivalence.at(i);
			}
			of << " )" <<std::endl;
		}
}

MonteCarloNode::MonteCarloNode(const std::weak_ptr<MonteCarloNode>& parent, PartitionManager&& hp, const History& new_history)
	: tree_(parent.lock()->tree_), parent_(parent), partitions_(std::move(hp)), history_(nullptr) {

	calculateHope();

	for(const auto &eq : new_history.getChoices()) {
		double result = 1;
		for(size_t i = 0; i < eq.size(); i++) {
			result *= PartitionManager::EQUIVALENCE_MODIFIER;
		}
		fact_ += result;
	}

	fact_ *= FACT_MODIFIER;
	weight_ = hope_ + fact_;
}

MonteCarloNode::MonteCarloNode(PartitionManager&& hp, MonteCarloTree* tree)
	: tree_(tree), partitions_(std::move(hp)), history_(nullptr) {
	calculateHope();
	weight_ = hope_;
}

MonteCarloNode::MonteCarloNode(MonteCarloNode&& other) noexcept : hope_(other.hope_), fact_(other.fact_), weight_(other.weight_),
tree_(other.tree_), parent_(std::move(other.parent_)), children_(std::move(other.children_)),
partitions_(std::move(other.partitions_)), history_(other.history_) {};

MonteCarloNode& MonteCarloNode::operator=(MonteCarloNode&& other) noexcept {
	hope_ = other.hope_;
	fact_ = other.fact_;
	weight_ = other.weight_;
	tree_ = other.tree_;
	parent_ = std::move(other.parent_);
	children_ = std::move(other.children_);
	partitions_ = std::move(other.partitions_);
	history_ = other.history_;
	return *this;
}

void MonteCarloNode::backpropagate() const {
	const double p_hope = parent_.lock()->hope_;
	const double hope_change = hope_ - p_hope;
	if (hope_change < 0)
		propagate(MonteCarloNode::PROPAGATION_INITIAL_VALUE, hope_change);
}


void MonteCarloNode::propagate(const double k, const double value) const {
	if(auto ptr = parent_.lock()) {
		std::weak_ptr<MonteCarloNode> weak_ptr = ptr;
		if(ptr->fact_ == 0) {
			ptr->updateHopeAndWeight(k * value, weak_ptr);
			return;
		}
		ptr->updateHopeAndWeight(k*value, weak_ptr);
		ptr->propagate(k * PROPAGATION_MODIFIER, value);
	}
}

void MonteCarloNode::updateHopeAndWeight(const double value, std::weak_ptr<MonteCarloNode>& weak_ptr) {
	const double old_weight = weight_;
	if (hope_ + value > 0) {
		hope_ += value;
		weight_ += value;
	} else {
		hope_ = 0;
		weight_ = fact_;
	}
	if (tree_->isInCandidates(weak_ptr))
		tree_->updateCandidatesTotal(weight_ - old_weight);
}

void MonteCarloNode::calculateHope() {

	size_t max_size = 0;
	std::vector<std::vector<size_t>> all_chain_lengths;
	for(size_t i = 0; i < partitions_.size(); i++) {
		std::vector<size_t> chain_lengths;
		for(const auto& node : partitions_.getHp().at(i).partition_heads) {
			auto dim = partitions_.chainLength(node);
			if(dim > 0) {
				chain_lengths.push_back(dim);
			}
		}
		if(chain_lengths.size() < 2)
			continue;
		std::sort(chain_lengths.begin(), chain_lengths.end(),
				  [](const size_t& a, const size_t& b) {return a < b; });
		if(chain_lengths.size() > max_size)
			max_size = chain_lengths.size();
		all_chain_lengths.push_back(std::move(chain_lengths));
	}

	std::vector<size_t> n_best_case_eqs(max_size, 0);

	for(auto& curr_length : all_chain_lengths) {
		size_t min_dim = curr_length.front();
		size_t curr_size = curr_length.size();
		size_t index = 1;
		{
			while(min_dim > 0 && curr_size > 1) {
				n_best_case_eqs.at(curr_size - 2) += min_dim;

				std::transform(curr_length.begin() + index, curr_length.end(), curr_length.begin() + index,
							   [min_dim](size_t d) -> std::size_t { return d - min_dim; });

				min_dim = curr_length.at(index);
				index++;
				curr_size--;
			}
		}

	}

	double modifier = 1;

	for(size_t eq_size = 0; eq_size < n_best_case_eqs.size(); eq_size++) {
		for(size_t k = 0; k < n_best_case_eqs.at(eq_size); k++) {
			double result = 1;
			for(size_t j = 0; j < eq_size + 2; j++) {
				result *= PartitionManager::EQUIVALENCE_MODIFIER;
			}
			hope_ += result * modifier;
			modifier *= PartitionManager::HOPE_REDUCER;
		}
	}

	hope_ *= NODE_HOPE_MODIFIER;
}


