//GAMY: a tool for estimating the similarity of graphs
//	Copyright(C) 2019 Mattia De Prisco and Giovanni Squillero
//
//	This program is free software : you can redistribute it and /or modify
//	it under the terms of the GNU General Public License as published by
//	the Free Software Foundation, either version 3 of the License, or
//	(at your option) any later version.
//
//	This program is distributed in the hope that it will be useful,
//	but WITHOUT ANY WARRANTY; without even the implied warranty of
//	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.See the
//	GNU General Public License for more details.
//
//	You should have received a copy of the GNU General Public License
//	along with this program.If not, see < http://www.gnu.org/licenses/>.

#pragma once

#include "hyperpartition.h"
#include "history.h"
#include "selector.h"
#include <vector>
#include <memory>

class MonteCarloTree;

class MonteCarloNode {

	double hope_ = 0;
	double fact_ = 0;
	double weight_ = 0;

	MonteCarloTree* tree_;
	std::weak_ptr<MonteCarloNode> parent_;
	std::vector<std::weak_ptr<MonteCarloNode>> children_;

	PartitionManager partitions_;   // nb: default constructor is automagically deleted as Hyper_partition is not "default constructible"
	const History* history_;

	double partitionValue(const size_t partition_index) const { return partitions_.getHp()[partition_index].partition_hope; };
	void propagate(const double k, const double value) const;
	void updateHopeAndWeight(const double value, std::weak_ptr<MonteCarloNode>& weak_ptr);
	void calculateHope();

public:
	static double NODE_HOPE_MODIFIER;
	static double FACT_MODIFIER;
	static double PROPAGATION_INITIAL_VALUE;
	static double PROPAGATION_MODIFIER;

	MonteCarloNode() = delete;
	~MonteCarloNode() = default;
	MonteCarloNode(const std::weak_ptr<MonteCarloNode>& parent, PartitionManager&& hp, const History& new_history);
	MonteCarloNode(PartitionManager&& hp, MonteCarloTree* tree);

	MonteCarloNode(const MonteCarloNode& other) = delete;
	MonteCarloNode& operator=(const MonteCarloNode& other) = delete;
	MonteCarloNode(MonteCarloNode&& other) noexcept;
	MonteCarloNode& operator=(MonteCarloNode&& other) noexcept;

	void backpropagate() const;
	std::vector<std::weak_ptr<MonteCarloNode>>& getChildren() { return children_; }
	std::weak_ptr<MonteCarloNode> getParent() const { return parent_; }
	double getFact() const { return fact_; }
	double getWeight() const { return weight_; };
	const History& getHistory() const { return *history_; }

	void updateParent(const std::weak_ptr<MonteCarloNode>& new_parent) { parent_ = new_parent; }
	void setHistory(const History* history) { history_ = history; }

	bool findNewEquivalence(const std::weak_ptr<MonteCarloNode>& weak_this);
	void expand(History&& new_history, const std::vector<node_t>& new_equivalence, const std::weak_ptr<MonteCarloNode>& weak_this);

	void printEquivalences(std::ofstream& of) const;
};