//GAMY: a tool for estimating the similarity of graphs
//	Copyright(C) 2019 Mattia De Prisco and Giovanni Squillero
//
//	This program is free software : you can redistribute it and /or modify
//	it under the terms of the GNU General Public License as published by
//	the Free Software Foundation, either version 3 of the License, or
//	(at your option) any later version.
//
//	This program is distributed in the hope that it will be useful,
//	but WITHOUT ANY WARRANTY; without even the implied warranty of
//	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.See the
//	GNU General Public License for more details.
//
//	You should have received a copy of the GNU General Public License
//	along with this program.If not, see < http://www.gnu.org/licenses/>.

#include "montecarlotree.h"
#include "utils.h"
#include <mutex>
#include <condition_variable>
#include <fstream>

unsigned int MonteCarloTree::MIN_CANDIDATES = 1;
unsigned int MonteCarloTree::CYCLES = 0;
std::string MonteCarloTree::RESULT_FOLDER = "../results/";
double MonteCarloTree::BAD_WEIGHT_RATIO = 1.0;
unsigned int MonteCarloTree::TIMEOUT = 10;
std::atomic<bool> MonteCarloTree::TIMED_OUT;

void MonteCarloTree::MonteCarloTreeSearch(PartitionManager&& initial_hyper_partition, std::vector<Graph>&& graphs) {

	TIMED_OUT.store(false);

	std::thread timeout_thread;
	std::mutex timeout_mutex;
	std::condition_variable timeout_cv;

	if (TIMEOUT != 0) {
		timeout_thread = std::thread([&] {
			const auto stop_time = std::chrono::steady_clock::now() + std::chrono::seconds(TIMEOUT);
			{
				std::unique_lock<std::mutex> guard(timeout_mutex);
				if(std::cv_status::timeout == timeout_cv.wait_until(guard, stop_time))
					TIMED_OUT.store(true);
			}
			});
	}

	MonteCarloTree tree{ std::move(initial_hyper_partition), std::move(graphs) };

	bool exausthed = false;
	try {
		while (!exausthed && !TIMED_OUT) {
			CYCLES++;
			tree.helper.updateRamUsage();
			exausthed = !tree.expand();
		}
	} catch(...) {
		std::ofstream failed_tests("../failed_tests", std::ios_base::app);
		tree.helper.printInput(failed_tests);
		failed_tests << "-------------------------------------------" << std::endl;

		if (timeout_thread.joinable()) {
			{
				std::unique_lock<std::mutex> guard(timeout_mutex);
				timeout_cv.notify_all();
			}
			timeout_thread.join();
		}
		return;
	}

	const std::clock_t time_elapsed = std::clock() - tree.helper.start_wall_clock_time;
	const auto cpu_time_elapsed = std::chrono::duration_cast<std::chrono::milliseconds>
		(std::chrono::steady_clock::now() - tree.helper.start_cpu_time).count();

	if (timeout_thread.joinable()) {
		{
			std::unique_lock<std::mutex> guard(timeout_mutex);
			timeout_cv.notify_all();
		}
		timeout_thread.join();
	}

	tree.candidates_.clear();
	const auto graphs_equivalences = tree.getGraphsEqs();
	const auto best_eq_dims = tree.getBestEqsDimensions();
	const auto equivalence_index = tree.computeEquivalenceIndex();
	tree.helper.printResults(time_elapsed, cpu_time_elapsed, exausthed, graphs_equivalences, best_eq_dims, equivalence_index);
}

// Choose a node in the tree to be expanded. 
bool MonteCarloTree::expand() {
	while (candidates_.size() > 0) {
		auto selected = candidates_.getCandidate();
		if (auto node = selected.first.lock()) {
			if (!badWeight(node->getWeight(), node->getFact(), false)) {
				if (node->findNewEquivalence(selected.first))
					return true;
				else
					candidates_.rmvEntry(selected.second);
			} else {
				candidates_.rmvEntry(selected.second);
				closeNode(node->getHistory());
				helper.closed_old++;
			}
		}
	}
	return false;
}

std::vector<int> MonteCarloTree::getBestEqsDimensions() const {
	std::vector<int> dims(graphs_.size());
	for (const auto& eq : best_solution_.lock()->getHistory().getChoices()) {
		dims.at(eq.size() - 1)++;
	}
	return dims;
}

unsigned MonteCarloTree::getNumSolutions() const {
	unsigned int total = 0;
	for (const auto& n : nodes_)
		if (n.second != nullptr && n.second->getFact() != 0)
			total++;

	return total;
}

void MonteCarloTree::updateBestSolution(const std::shared_ptr<MonteCarloNode>& new_node) {
	if (new_node->getFact() <= best_fact_)
		return;

	helper.best_wall_clock_time = std::clock() - helper.start_wall_clock_time;
	helper.best_cpu_time = std::chrono::duration_cast<std::chrono::milliseconds>
		(std::chrono::steady_clock::now() - helper.start_cpu_time).count();

	best_fact_ = new_node->getFact();
	best_solution_ = new_node;
}

bool MonteCarloTree::badWeight(const double weight, const double fact, bool new_node) {
	if(!new_node) {
		if (fact == 0 || fact == best_fact_)
			return false;
		return candidates_.size() >= MIN_CANDIDATES && weight * BAD_WEIGHT_RATIO < best_fact_;
	}

	if (candidates_.size() < MIN_CANDIDATES || fact == best_fact_)
		return false;
	if(weight * BAD_WEIGHT_RATIO < best_fact_) {
		const auto worst_candidate = candidates_.getWorstCandidate();
		if(worst_candidate.first.lock()->getWeight() < weight) {
			candidates_.rmvEntry(worst_candidate.second);
			closeNode(worst_candidate.first.lock()->getHistory());
			helper.closed_old++;
			return false;
		}
		return true;
	}
	return false;
}

void MonteCarloTree::insertNewNode(History&& history, std::shared_ptr<MonteCarloNode>&& node) {
	auto new_node = nodes_.insert(std::make_pair(std::move(history), std::move(node)));
	new_node.first->second->setHistory(&new_node.first->first);
}


MonteCarloTree::MonteCarloTree(PartitionManager&& hyper_partition, std::vector<Graph>&& graphs) : graphs_(std::move(graphs)) {
	auto root_node = std::make_shared<MonteCarloNode>(std::move(hyper_partition), this);
	root_ = root_node;
	std::weak_ptr<MonteCarloNode> ptr = root_node;
	insertNewCandidate(std::move(ptr));
	insertNewNode(History(), std::move(root_node));
	helper = MonteCarloTreeHelper(this);
}


void MonteCarloTree::insertNewCandidate(std::weak_ptr<MonteCarloNode>&& node) {
	if (!node.expired())
		candidates_.addEntry(std::move(node));
}

void MonteCarloTree::closeNode(const History& history) {
	auto it = nodes_.find(history);
	for (auto& child : it->second->getChildren()) {
		if (auto ptr = child.lock()) {
			ptr->updateParent(it->second->getParent());
		}
	}
	nodes_.erase(it);
}

std::unordered_map<uint64_t, int> MonteCarloTree::getGraphsEqs() const {
	std::unordered_map<uint64_t, int> graphs_equivalences;

	for (auto& solution : nodes_) {
		if (solution.second == nullptr || solution.second->getFact() == 0)
			continue;

		std::unordered_map<uint64_t, int> stat_map;

		//Inserting one entry per different pair of graphs in the map,
		//each one indicating the number of nodes they have in common
		for (size_t i = 0; i < graphs_.size(); i++) {
			for (size_t j = i + 1; j < graphs_.size(); j++) {
				stat_map.insert(std::make_pair(Pair(static_cast<uint32_t>(i), static_cast<uint32_t>(j)), 0));
			}
		}

		for (const auto& equivalence : solution.second->getHistory().getChoices()) {
			if (equivalence.size() == graphs_.size()) {
				for (auto& common_nodes : stat_map) {
					common_nodes.second++;
				}
			} else {
				std::vector<size_t> graph_present(graphs_.size(), 0);
				size_t current = graphs_.at(static_cast<int>(0)).size();
				int current_graph = 0;

				for (const auto& graph_node : equivalence) {
					while (true) {
						if (graph_node < static_cast<int>(current)) {
							graph_present.at(current_graph)++;
							break;
						}
						current_graph++;
						current += graphs_.at(current_graph).size();
					}
				}

				for (size_t i = 0; i < graph_present.size(); i++) {
					if (graph_present.at(i) == 1) {
						for (size_t j = i + 1; j < graph_present.size(); j++) {
							if (graph_present.at(j) == 1) {
								stat_map.find(Pair(static_cast<uint32_t>(i), static_cast<uint32_t>(j)))->second++;
							}
						}
					}
				}
			}
		}
		for (const auto& current_eq : stat_map) {
			auto graph_equivalence = graphs_equivalences.find(current_eq.first);
			if (graph_equivalence == graphs_equivalences.end()) {
				graphs_equivalences.insert(current_eq);
			} else if (graph_equivalence->second < current_eq.second) {
				graph_equivalence->second = current_eq.second;
			}
		}
	}
	return graphs_equivalences;
}

double MonteCarloTree::computeEquivalenceIndex() const {
	double max_index = 0;
	std::vector<int> ideal_equivalences;
	ideal_equivalences.reserve(graphs_.size());
	auto graphs_copy = graphs_;
	std::sort(graphs_copy.begin(), graphs_copy.end(), [](const Graph& g0, const Graph& g1) {return g0.size() < g1.size(); });

	size_t last_size = 0;
	for (size_t i = 0; i < graphs_copy.size() - 1; i++) {
		const size_t n_equivalences = graphs_copy.at(i).size() - last_size;
		ideal_equivalences.push_back(n_equivalences);
		last_size += n_equivalences;
	}

	std::reverse(ideal_equivalences.begin(), ideal_equivalences.end());

	for(size_t eq_length = 0; eq_length < ideal_equivalences.size(); eq_length++) {
		double value = 1;
		for(size_t i = 0; i < eq_length + 2; i++) {
			value *= PartitionManager::EQUIVALENCE_MODIFIER;
		}
		max_index += value * ideal_equivalences.at(eq_length);
	}

	return best_fact_ / MonteCarloNode::FACT_MODIFIER / max_index;
}

MonteCarloTreeHelper::MonteCarloTreeHelper(MonteCarloTree* tree) : tree_(tree) {
	start_wall_clock_time = std::clock();
	start_cpu_time = std::chrono::steady_clock::now();
}

void MonteCarloTreeHelper::updateRamUsage() {
#if defined(ON_WINDOWS)
	PROCESS_MEMORY_COUNTERS_EX pmc;
	GetProcessMemoryInfo(GetCurrentProcess(), reinterpret_cast<PROCESS_MEMORY_COUNTERS*>(&pmc), sizeof(pmc));
	const float current_usage = static_cast<float>(pmc.WorkingSetSize) / 1024.0;
	memory_counter++;
	memory_usage_total += current_usage;
	if (current_usage > memory_usage_peak)
		memory_usage_peak = current_usage;
#else
	FILE* file = fopen("/proc/self/status", "r");
	int current_usage = -1;
	char line[128];

	while (fgets(line, 128, file) != NULL) {
		if (strncmp(line, "VmRSS:", 6) == 0) {
			current_usage = ParseLine(line);
			break;
		}
	}
	fclose(file);
	memory_counter++;
	memory_usage_total += current_usage;
	if (current_usage > memory_usage_peak)
		memory_usage_peak = current_usage;
#endif
}


std::string MonteCarloTreeHelper::computeFilename() const {
	std::string	filename(MonteCarloTree::RESULT_FOLDER);

	for (const auto& graph : tree_->graphs_) {
		filename += graph.getFileName() + "_";
	}

	const auto attach_double_str_without_zeros = [](std::string& base, std::string str) {
		str.erase(str.find_last_not_of('0') + 1, std::string::npos);
		if (str.back() == '.')
			str.erase(str.end() - 1);
		base += str;
	};

	filename += "t" + std::to_string(MonteCarloTree::TIMEOUT);
	attach_double_str_without_zeros(filename, "_em" + std::to_string(PartitionManager::EQUIVALENCE_MODIFIER));
	attach_double_str_without_zeros(filename, "_hr" + std::to_string(PartitionManager::HOPE_REDUCER));
	attach_double_str_without_zeros(filename, "_nhm" + std::to_string(MonteCarloNode::NODE_HOPE_MODIFIER));
	attach_double_str_without_zeros(filename, "_fm" + std::to_string(MonteCarloNode::FACT_MODIFIER));
	attach_double_str_without_zeros(filename, "_bwr" + std::to_string(MonteCarloTree::BAD_WEIGHT_RATIO));
	attach_double_str_without_zeros(filename, "_piv" + std::to_string(MonteCarloNode::PROPAGATION_INITIAL_VALUE));
	attach_double_str_without_zeros(filename, "_pm" + std::to_string(MonteCarloNode::PROPAGATION_MODIFIER));
	filename += "_mc" + std::to_string(MonteCarloTree::MIN_CANDIDATES);
	filename += "_l" + std::to_string(Graph::VERTEX_LABELS_ON);

	return filename;
}

void MonteCarloTreeHelper::printResults(const clock_t time_elapsed, const long long cpu_time_elapsed, const bool exausthed,
										const std::unordered_map<uint64_t, int>& graphs_equivalences, 
										const std::vector<int>& best_eq_dims, const double equivalence_index) const {

	const std::string filename = computeFilename();

	std::ofstream output_file;
	output_file.open(filename, std::ios::out);

	printInput(output_file);
	printTime(output_file, time_elapsed, cpu_time_elapsed);
	printRamUsage(output_file);
	printStats(output_file, exausthed);

	printEquivalenceIndex(output_file, equivalence_index);
	if (equivalence_index != 0) {
		printGraphsEquivalences(output_file, graphs_equivalences);
		printBestSolutionStats(output_file, best_eq_dims);
		printBestSolutionEqs(output_file);
	}
	output_file.close();
}

void MonteCarloTreeHelper::printBestStats(const std::clock_t time_elapsed, const std::unordered_map<uint64_t, int>& graphs_equivalences,
											const std::vector<int>& best_eq_dims) const {
	const std::string filename = computeFilename();
	std::ofstream output_file;
	output_file.open(filename, std::ios::app);
	printTimeAndCycles(output_file, time_elapsed);
	printStats(output_file, false);
	printBestSolutionStats(output_file, best_eq_dims);
	printGraphsEquivalences(output_file, graphs_equivalences);
	output_file.close();
}

void MonteCarloTreeHelper::printTimeAndCycles(std::ofstream& of, const clock_t time_elapsed) const {
	of << "########################################################################################" << std::endl;
	of << "New best at cycle " << MonteCarloTree::CYCLES << ", at " << time_elapsed * 1000 / CLOCKS_PER_SEC << " ms time mark" << std::endl;
	of << std::endl;
}

void MonteCarloTreeHelper::printExhaustion() const {
	const std::string filename = computeFilename();
	std::ofstream output_file;
	output_file.open(filename, std::ios::app);
	output_file << "************** EXITED FOR EXHAUSTION **************" << std::endl;
	output_file.close();
}

void MonteCarloTreeHelper::printHeader() const {
	const std::string filename = computeFilename();
	std::ofstream output_file;
	output_file.open(filename, std::ios::out);
	printInput(output_file);
	output_file.close();
}

void MonteCarloTreeHelper::printInput(std::ofstream& of) const {
	of << "### Input graphs ###" << std::endl;
	for (size_t i = 0; i < tree_->graphs_.size(); i++)
		of << "Graph " << i << " \""
		<< tree_->graphs_.at(i).getApplicationName()
		<< "\" (" << tree_->graphs_.at(i).getFileName() << "), with "
		<< tree_->graphs_.at(i).size() << " nodes" << std::endl;

	of << std::endl;

	of << "### Input parameters ###" << std::endl;
	of << "Timeout: " << MonteCarloTree::TIMEOUT << std::endl;
	of << "Equivalence Modifier: " << PartitionManager::EQUIVALENCE_MODIFIER << std::endl;
	of << "Partition Hope Reducer: " << PartitionManager::HOPE_REDUCER << std::endl;
	of << "Monte Carlo Node Hope Modifier: " << MonteCarloNode::NODE_HOPE_MODIFIER << std::endl;
	of << "Fact Modifier: " << MonteCarloNode::FACT_MODIFIER << std::endl;
	of << "Bad Weight Ratio: " << MonteCarloTree::BAD_WEIGHT_RATIO << std::endl;
	of << "Propagation Initial Value: " << MonteCarloNode::PROPAGATION_INITIAL_VALUE << std::endl;
	of << "Propagation Modifier: " << MonteCarloNode::PROPAGATION_MODIFIER << std::endl;
	of << "Minimum candidates: " << MonteCarloTree::MIN_CANDIDATES << std::endl;
	of << "Vertex labels: ";
	if (Graph::VERTEX_LABELS_ON) {
		of << "ON";
	}
	else {
		of << "OFF";
	}
	of << std::endl;

	of << std::endl;
}

void MonteCarloTreeHelper::printTime(std::ofstream& of, const clock_t time_elapsed, const long long cpu_time_elapsed) const {
	of << "### Time ###" << std::endl;
	of << "Total MCTS wall-clock time (ms): " << time_elapsed * 1000 / CLOCKS_PER_SEC << std::endl;
	of << "Wall-clock time to find best solution (ms): " << best_wall_clock_time * 1000 / CLOCKS_PER_SEC << std::endl;

	of << std::endl;

	of << "Total MCTS CPU time (ms): " << cpu_time_elapsed << std::endl;
	of << "CPU time to find best solution (ms): " << best_cpu_time << std::endl;

	of << std::endl;
}

void MonteCarloTreeHelper::printRamUsage(std::ofstream& of) const {
	of << "### RAM Usage ###" << std::endl;
	of << "Average RAM usage sampled at every cycle (KB): " << memory_usage_total / memory_counter << std::endl;
	of << "RAM usage peak (KB): " << memory_usage_peak << std::endl;

	of << std::endl;
}

void MonteCarloTreeHelper::printStats(std::ofstream& of, const bool exausthed) const {
	of << "### Statistics ###" << std::endl;
	if (exausthed)
		of << "Exited for exhaustion." << std::endl;
	else if (MonteCarloTree::TIMED_OUT)
		of << "Exited for timeout." << std::endl;

	of << "MCTS cycles: " << MonteCarloTree::CYCLES << std::endl;
	of << "Same choices avoided: " << avoided << std::endl;
	of << "Nodes closed by exhaustion: " << closed_by_exausthion << std::endl;
	of << "New nodes closed by bad weight: " << closed_new << std::endl;
	of << "Old nodes closed by bad weight: " << closed_old << std::endl;
	of << "Final solution space size: " << tree_->getNumSolutions() << std::endl;
	of << std::endl;
}

void MonteCarloTreeHelper::printEquivalenceIndex(std::ofstream& of, const double equivalence_index) const {
	of << "### RESULTS ###" << std::endl;
	if(equivalence_index == 0) {
		of << "No solutions found" << std::endl;
		return;
	}
	of << "Equivalence index: " << equivalence_index << std::endl;
	of << std::endl;
}

void MonteCarloTreeHelper::printGraphsEquivalences(std::ofstream& of, const std::unordered_map<uint64_t, int>& graphs_equivalences) const {
	of << "Specific equivalences: " << std::endl;
	for (auto& s : graphs_equivalences) {
		uint32_t first, second;
		Depair(s.first, first, second);
		of << "- Graph " << first << " is "
			<< s.second / static_cast<double>(tree_->graphs_.at(first).size()) * 100
			<< "% equal to graph " << second << std::endl;
		of << "- Graph " << second << " is "
			<< s.second / static_cast<double>(tree_->graphs_.at(second).size()) * 100
			<< "% equal to graph " << first << std::endl;
	}
	of << std::endl;
}

void MonteCarloTreeHelper::printBestSolutionStats(std::ofstream& of, const std::vector<int>& best_eq_dims) const {
	of << "Best solution: " << std::endl;
	for (size_t i = tree_->graphs_.size() - 1; i > 0; --i) {
		of << "- Equivalences with " << i + 1 << " nodes: " << best_eq_dims.at(i) << std::endl;
	}
	of << std::endl;
}

void MonteCarloTreeHelper::printBestSolutionEqs(std::ofstream& of) const {
	of << "### BEST SOLUTION EQUIVALENCES ###" << std::endl;
	tree_->best_solution_.lock()->printEquivalences(of);
	of << std::endl;
}

bool MonteCarloTreeHelper::checkCorrectness() const {
	for (const auto& equivalence : tree_->best_solution_.lock()->getHistory().getChoices()) {
		auto label = NodeStore::getNode(equivalence.front())->label;
		std::vector<node_t> from, to;
		for (const auto& node : equivalence) {
			if (label != NodeStore::getNode(node)->label)
				return false;

			for (auto& from_node : NodeStore::getNode(node)->from()) {
				from.push_back(from_node);
			}
			for (auto& to_node : NodeStore::getNode(node)->to()) {
				to.push_back(to_node);
			}
		}
		for (const auto& fn : from) {
			for (const auto& eq : tree_->best_solution_.lock()->getHistory().getChoices()) {
				bool found = false;
				for (const auto& n : eq) {
					if (n == fn) {
						found = true;
						break;
					}
				}
				if (found) {
					for (const auto& n : eq) {
						if (std::find(from.begin(), from.end(), n) == from.end()) {
							return false;
						}
					}
				}
			}
		}
		for (const auto& tn : to) {
			for (const auto& eq : tree_->best_solution_.lock()->getHistory().getChoices()) {
				bool found = false;
				for (const auto& n : eq) {
					if (n == tn) {
						found = true;
						break;
					}
				}
				if (found) {
					for (const auto& n : eq) {
						if (std::find(to.begin(), to.end(), n) == to.end()) {
							return false;
						}
					}
				}
			}
		}
	}
	return true;
}
