//GAMY: a tool for estimating the similarity of graphs
//	Copyright(C) 2019 Mattia De Prisco and Giovanni Squillero
//
//	This program is free software : you can redistribute it and /or modify
//	it under the terms of the GNU General Public License as published by
//	the Free Software Foundation, either version 3 of the License, or
//	(at your option) any later version.
//
//	This program is distributed in the hope that it will be useful,
//	but WITHOUT ANY WARRANTY; without even the implied warranty of
//	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.See the
//	GNU General Public License for more details.
//
//	You should have received a copy of the GNU General Public License
//	along with this program.If not, see < http://www.gnu.org/licenses/>.

#pragma once

#include <chrono>
#include <vector>
#include <ctime>
#include <memory>
#include "montecarlonode.h"
#include "nodeselector.h"
#include "graph.h"
#include <algorithm>
#include <unordered_map>
#include <atomic>

#if defined(WIN32) || defined(_WIN32) || defined(__WIN32) && !defined(__CYGWIN__)
#define ON_WINDOWS
#include <Windows.h>
#include <Psapi.h>
#else
#include <string.h>
#endif

class MonteCarloNode;

class MonteCarloTreeHelper {
public:

	MonteCarloTree* tree_;

	int closed_by_exausthion = 0;
	int closed_new = 0;
	int closed_old = 0;
	int avoided = 0;

	std::clock_t start_wall_clock_time;
	std::clock_t best_wall_clock_time;
	std::chrono::steady_clock::time_point start_cpu_time;
	long long best_cpu_time;

	float memory_usage_peak = 0;
	float memory_usage_total = 0;
	size_t memory_counter = 0;

	MonteCarloTreeHelper() = default;
	explicit MonteCarloTreeHelper(MonteCarloTree* tree);

	void updateRamUsage();
	std::string computeFilename() const;
	void printResults(const clock_t time_elapsed, const long long cpu_time_elapsed, const bool exausthed,
				const std::unordered_map<uint64_t, int>& graphs_equivalences,
				const std::vector<int>& best_eq_dims, const double equivalence_index) const;
	void printInput(std::ofstream& of) const;
	void printTime(std::ofstream& of, const clock_t time_elapsed, const long long cpu_time_elapsed) const;
	void printRamUsage(std::ofstream& of) const;
	void printStats(std::ofstream& of, const bool exausthed) const;
	void printEquivalenceIndex(std::ofstream& of, const double equivalence_index) const;
	void printGraphsEquivalences(std::ofstream& of, const std::unordered_map<uint64_t, int>& graphs_equivalences) const;
	void printBestSolutionStats(std::ofstream& of, const std::vector<int>& best_eq_dims) const;
	void printBestSolutionEqs(std::ofstream& of) const;

	void printHeader() const;
	void printBestStats(const std::clock_t time_elapsed, const std::unordered_map<uint64_t, int>& graphs_equivalences,
						const std::vector<int>& best_eq_dims) const;
	void printTimeAndCycles(std::ofstream& of, const clock_t time_elapsed) const;
	void printExhaustion() const;

	bool checkCorrectness() const;
};


class MonteCarloTree {
	friend class MonteCarloTreeHelper;

	std::vector<Graph> graphs_;
	std::weak_ptr<MonteCarloNode> root_;
	std::unordered_map<History, std::shared_ptr<MonteCarloNode>> nodes_;
	NodeSelector candidates_;
	std::weak_ptr<MonteCarloNode> best_solution_;
	double best_fact_ = 0;

public:
	MonteCarloTreeHelper helper;

	static unsigned int MIN_CANDIDATES;
	static unsigned int CYCLES;
	static double BAD_WEIGHT_RATIO;
	static std::string RESULT_FOLDER;
	static unsigned int TIMEOUT;
	static std::atomic<bool> TIMED_OUT;

	explicit MonteCarloTree(PartitionManager&& hyper_partition, std::vector<Graph>&& graphs);
	void insertNewCandidate(std::weak_ptr<MonteCarloNode>&& node);
	void updateCandidatesTotal(const double value) { candidates_.updateTotal(value); }
	void closeNode(const History& history);
	bool findNode(const History& history) { return nodes_.find(history) != nodes_.end(); }
	bool isInCandidates(std::weak_ptr<MonteCarloNode>& node) { return candidates_.contains(node); };
	std::shared_ptr<MonteCarloNode>& getNode(const History& history) { return nodes_.find(history)->second; };
	void updateBestSolution(const std::shared_ptr<MonteCarloNode>& new_node);
	bool badWeight(const double weight, const double fact, bool new_node);
	void insertNewNode(History&& history, std::shared_ptr<MonteCarloNode>&& node);
	void insertNullNode(History&& history) { nodes_.insert(std::make_pair(std::move(history), nullptr)); }

	bool expand();
	std::vector<int> getBestEqsDimensions() const;
	unsigned int getNumSolutions() const;
	std::unordered_map<uint64_t, int> getGraphsEqs() const;
	double computeEquivalenceIndex() const;
	
	static void MonteCarloTreeSearch(PartitionManager&& initial_hyper_partition, std::vector<Graph>&& graphs);
};