//GAMY: a tool for estimating the similarity of graphs
//	Copyright(C) 2019 Mattia De Prisco and Giovanni Squillero
//
//	This program is free software : you can redistribute it and /or modify
//	it under the terms of the GNU General Public License as published by
//	the Free Software Foundation, either version 3 of the License, or
//	(at your option) any later version.
//
//	This program is distributed in the hope that it will be useful,
//	but WITHOUT ANY WARRANTY; without even the implied warranty of
//	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.See the
//	GNU General Public License for more details.
//
//	You should have received a copy of the GNU General Public License
//	along with this program.If not, see < http://www.gnu.org/licenses/>.

#include "nodeselector.h"
#include "montecarlotree.h"

std::pair<std::weak_ptr<MonteCarloNode>, size_t> NodeSelector::getWorstCandidate() {

	auto itr = std::min_element(entries_.begin(), entries_.end(), []
	(const std::weak_ptr<MonteCarloNode>& entry_a, const std::weak_ptr<MonteCarloNode>& entry_b) {
			return entry_a.lock()->getWeight() < entry_b.lock()->getWeight();
	});
	return std::make_pair(*itr, itr - entries_.begin());
}

std::pair<std::weak_ptr<MonteCarloNode>, size_t > NodeSelector::getCandidate() {
	if (entries_.size() == 1)
		return std::make_pair(entries_.front(), 0);
	return rouletteWheel();
}

bool NodeSelector::contains(std::weak_ptr<MonteCarloNode>& node_to_find) {
	return std::find_if(
		entries_.begin(), entries_.end(),
		[&node_to_find](const std::weak_ptr<MonteCarloNode>& n) {
		return node_to_find.lock() == n.lock();
	}) != entries_.end();
}

void NodeSelector::addEntry(std::weak_ptr<MonteCarloNode>&& entry) {
	const double weight = entry.lock()->getWeight();
	entries_.push_back(std::move(entry));
	updateTotal(weight);
}

void NodeSelector::rmvEntry(const size_t pos) {
	const auto it = entries_.begin() + pos;
	updateTotal(-it->lock()->getWeight());
	entries_.erase(it);
}

double NodeSelector::getNormValue(const size_t pos) const {
	return entries_.at(pos).lock()->getWeight() / total_;
}

std::pair<std::weak_ptr<MonteCarloNode>, size_t> NodeSelector::rouletteWheel() const {
	std::uniform_real_distribution<long double> urd;
	size_t i = 0;
	for (auto rnd = urd(shared_random_engine, urd.param()); rnd >= 0; ++i) {
		rnd -= getNormValue(i);
	}
	return std::make_pair(entries_[i - 1], i - 1);

}

void NodeSelector::sortByWeight() {
	std::sort(entries_.begin(), entries_.end(),
		[](const std::weak_ptr<MonteCarloNode>& n0, const std::weak_ptr<MonteCarloNode>& n1) {
			return n0.lock()->getWeight() > n1.lock()->getWeight();
		});
}

bool NodeSelector::isSorted() {
	for(size_t i = 0; i < entries_.size() - 1; i++) {
		if (entries_.at(i).lock()->getWeight() < entries_.at(i + 1).lock()->getWeight())
			return false;
	}
	return true;
}
