//GAMY: a tool for estimating the similarity of graphs
//	Copyright(C) 2019 Mattia De Prisco and Giovanni Squillero
//
//	This program is free software : you can redistribute it and /or modify
//	it under the terms of the GNU General Public License as published by
//	the Free Software Foundation, either version 3 of the License, or
//	(at your option) any later version.
//
//	This program is distributed in the hope that it will be useful,
//	but WITHOUT ANY WARRANTY; without even the implied warranty of
//	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.See the
//	GNU General Public License for more details.
//
//	You should have received a copy of the GNU General Public License
//	along with this program.If not, see < http://www.gnu.org/licenses/>.

/*-*
 * Supports the basic selection methodologies from EC and other utilities
 *-*/

#pragma once

#include <type_traits>
#include <vector>
#include <algorithm>
#include <utility>
#include <random>
#include <thread>
#include <cassert>

#define USE_FAST_RANDOM

#ifdef USE_FAST_RANDOM
// the oldest and bestknown pseudorandom number generator, probably enough.
extern std::minstd_rand shared_random_engine;
#else
// the famous 64-bit mersenne twister by matsumoto & nishimura... is it really needed?
extern std::mt19937_64 Shared_random_engine;
#endif // USE_FAST_RANDOM

template<typename T, typename W>
class Selector {
    static_assert(std::is_arithmetic<W>::value, "Only arithmetic types are supported");  // TODO: (excercise) use std::enable_if

    std::vector<std::pair<T, W>> entries_;

public:
    size_t size() const { return entries_.size(); };

    typename std::vector<std::pair<T, W>>::const_iterator begin() const { return entries_.cbegin(); }
    typename std::vector<std::pair<T, W>>::const_iterator end() const { return entries_.cend(); }

    size_t addEntry(std::pair<T, W> entry) { entries_.push_back(entry); return size(); };
    size_t addEntry(T entry, W weight) { return addEntry(std::make_pair(entry, weight)); };

    // Completely random
    T random() {
        std::uniform_int_distribution<size_t> uid(0, entries_.size() - 1);
        return entries_[uid(shared_random_engine)].first;
    }

    // Tournament selection
    T tournament(int tournament_size = 2) {
        std::uniform_int_distribution<size_t> uid(0, entries_.size() - 1);
        std::vector<std::pair<T, W>> candidates;
        while(tournament_size > 0) {
            candidates.push_back(entries_[uid(shared_random_engine)]);
            --tournament_size;
        }
        auto m = std::max_element(candidates.begin(), candidates.end(),
            [](const std::pair<T, W>& a, const std::pair<T, W>& b) { return a.second < b.second; });
        return m->first;
    }

    // Roulette wheel
    T rouletteWheel() {
        W sum = 0;
        for(const auto e : entries_)
            sum += e.second;
        std::vector<long double> p;
        for(const auto e : entries_)
            p.push_back(static_cast<long double>(e.second) / static_cast<long double>(sum));

        std::uniform_real_distribution<long double> urd;
        size_t i = 0;
        for(auto rnd = urd(shared_random_engine, urd.param()); rnd >= 0; ++i) {
            rnd -= p[i];
        }
		auto result = std::move(entries_[i - 1].first);
		entries_.erase(entries_.begin() + i - 1);
		return result;
    }

    // Roulette wheel on linearized fitness
    T linearizedRouletteWheel() {
        // not yet implmented, but scholars claim that's equivalent
        return tournament(2);
    }
};

struct SelectorUtilities {
    static inline size_t tournamentRange(size_t range_size, int tournament_size) {
        std::uniform_int_distribution<size_t> uid(0, range_size - 1);
        assert(tournament_size > 0);
        size_t r = uid(shared_random_engine);
        while(tournament_size > 1) {
            r = std::max(r, uid(shared_random_engine));
            --tournament_size;
        }
        return r;
    }
    static inline size_t tournamentRangeInverted(size_t range_size, int tournament_size) {
        std::uniform_int_distribution<size_t> uid(0, range_size - 1);
        assert(tournament_size > 0);
        size_t r = uid(shared_random_engine);
        while(tournament_size > 1) {
            r = std::min(r, uid(shared_random_engine));
            --tournament_size;
        }
        return r;
    }

};
